from ..plate import Plate


def add_volume(
    destination_plate: Plate,
    destination_well: str,
    volume: float,
    volume_id: str,
    volume_concentration: int = None,
):
    """Add volume to a well of a plate object

    Args:
        destination_plate (Plate): target plate in which volume will be added to,
        destination_well (str): target well in which volume will be added to,
        volume (float): volume of liquid to add to the specified well,
        volume_id (str): identifier of liquid to be added to specified well,
        volume_concentration (int): concentraion of liquid to be added to specified well,

    Raises:
        ValueError: if volume well added exceeds well capacity
    """
    if (
        destination_plate[destination_well]["total_volume"] + volume
        <= destination_plate.well_volume
    ):
        destination_plate.set_well_volume(
            destination_well,
            float(destination_plate[destination_well]["total_volume"] + volume),
        )
        if volume_id in destination_plate[destination_well]["composition"]:
            new_volume = (
                destination_plate[destination_well]["composition"][volume_id]["volume"]
                + volume
            )
            destination_plate.update_well_composition(
                destination_well, new_volume, volume_id
            )
        else:
            destination_plate.update_well_composition(
                destination_well, volume, volume_id
            )
    else:
        raise ValueError(
            f"attempted to exceed plate: {destination_plate.id} with capacity ({destination_plate.well_volume}) "
            f"by adding {volume} nanoliters to already present {destination_plate[destination_well]['total_volume']} "
            f"at plate location {destination_well}"
        )



import uuid
import string
from typing import Literal, TypedDict, List, Dict
import math


class Composition_Item(TypedDict):
    volume: int
    concentration: int


class Contents(TypedDict):
    id: str
    total_volume: int
    composition: dict[str, Composition_Item]


class Plate:
    """Class for Plate objects.

    Generate a structure to handle plates and contents of wells

    Attributes:
        size (int): Number of wells within the plate.
        id (str): Unique identification of the plate object.
        deadspace (int): Volume of unaccessable / unreachable liquid within each well
        rows (list): row labels within plate e.g. ['A','B'...]
        columns (list): column labels within plate e.g. [1,2,3,4...]
        well_volume (int): volume of liquid within each well
    """

    def __init__(
        self,
        size: int,
        id: str = uuid.uuid4(),
        deadspace: int = None,
        rows: int = None,
        columns: int = None,
        well_volume: int = None,
    ):
        """Initiation object of Plate objects.

        Plate can be initiated which just size without rows and columns if size
        is of regular 2:3 ratio one of [6, 24, 96, 384, 1536]

        Args:
            size (int): Number of wells within the plate.
            id (str): Unique identification of the plate object.
            deadspace (int): Volume of unaccessable / unreachable liquid within each well
            rows (int): number of rows within the plate
            columns (int): number of columns within the plate
            well_volume (int): volume of liquid within each well

        Raises:
            ValueError: If rows or columns is a not a postive integer, or if rows and columns are
                not supplied and size is not regular 2:3 ratio.
        """
        self.size = size
        self.id = id
        self.deadspace = deadspace
        self.rows = rows
        self.columns = columns
        self.well_volume = well_volume
        self._test_input_validity()
        if rows:
            self.rows = [i for i in string.ascii_uppercase[:rows]]
        else:
            self.rows = [
                i for i in string.ascii_uppercase[: int(math.sqrt(self.size * (2 / 3)))]
            ]
        if columns:
            self.columns = [i for i in range(1, columns + 1)]
        else:
            self.columns = [
                i for i in range(1, int(math.sqrt(self.size * (3 / 2))) + 1)
            ]
        self.wells = [
            letter + str(number) for number in self.columns for letter in self.rows
        ]
        self.contents: Contents = {
            key: {"id": key, "total_volume": 0.0, "composition": {}}
            for key in self.wells
        }

    def __getitem__(self, k):
        """Return e.g. well A1's dict when calling myplate["A1"]."""
        return self.contents[k]

    def set_well_id(self, well: str, id: str):
        """Set id of individual wells"""
        self.contents[well]["id"] = id

    def set_well_volume(self, well: str, volume: int):
        """Set volume of individual wells"""
        self.contents[well]["total_volume"] = volume

    def update_well_composition(
        self, well: str, volume: int, volume_id: str, volume_concentration: int = None
    ):
        """Update composition of individual wells"""
        self.contents[well]["composition"].update(
            {volume_id: {"volume": volume, "concentration": volume_concentration}}
        )

    def _test_input_validity(self):
        if (
            self.size not in [6, 24, 96, 384, 1536]
            and self.rows == None
            and self.columns == None
        ):
            raise ValueError(
                "Plate size must be one of [6, 24, 96, 384, 1536] if rows and columns arent specified"
            )
        if self.columns:
            if self.columns <= 0:
                raise ValueError("Please use positive integer for columns")
        if self.rows:
            if self.rows <= 0:
                raise ValueError("Please use positive integer for rows")



import pytest
from platemap import Plate


def test_build_6_well_plate():
    """
    Tests 6 plate well is built correctly

    """
    plate6 = Plate(size=6)
    assert plate6.size == 6
    assert plate6.rows == ["A", "B"]
    assert plate6.columns == [1, 2, 3]
    assert plate6.wells == ["A1", "B1", "A2", "B2", "A3", "B3"]
    assert plate6.well_volume == None
    assert plate6.deadspace == None


def test_build_12_well_plate():
    """
    Tests 6 plate well is built correctly

    """
    plate6 = Plate(size=12, rows=3, columns=4)
    assert plate6.size == 12
    assert plate6.rows == ["A", "B", "C"]
    assert plate6.columns == [1, 2, 3, 4]
    assert plate6.wells == [
        "A1",
        "B1",
        "C1",
        "A2",
        "B2",
        "C2",
        "A3",
        "B3",
        "C3",
        "A4",
        "B4",
        "C4",
    ]
    assert plate6.well_volume == None
    assert plate6.deadspace == None


def test_build_24_well_plate():
    """
    Tests 6 plate well is built correctly

    """
    plate6 = Plate(size=24)
    assert plate6.size == 24
    assert plate6.rows == ["A", "B", "C", "D"]
    assert plate6.columns == [1, 2, 3, 4, 5, 6]
    assert plate6.wells == [
        "A1",
        "B1",
        "C1",
        "D1",
        "A2",
        "B2",
        "C2",
        "D2",
        "A3",
        "B3",
        "C3",
        "D3",
        "A4",
        "B4",
        "C4",
        "D4",
        "A5",
        "B5",
        "C5",
        "D5",
        "A6",
        "B6",
        "C6",
        "D6",
    ]
    assert plate6.well_volume == None
    assert plate6.deadspace == None


def test_build_24_well_plate():
    """
    Tests 6 plate well is built correctly

    """
    plate6 = Plate(size=24)
    assert plate6.size == 24
    assert plate6.rows == ["A", "B", "C", "D"]
    assert plate6.columns == [1, 2, 3, 4, 5, 6]
    assert plate6.wells == [
        "A1",
        "B1",
        "C1",
        "D1",
        "A2",
        "B2",
        "C2",
        "D2",
        "A3",
        "B3",
        "C3",
        "D3",
        "A4",
        "B4",
        "C4",
        "D4",
        "A5",
        "B5",
        "C5",
        "D5",
        "A6",
        "B6",
        "C6",
        "D6",
    ]
    assert plate6.well_volume == None
    assert plate6.deadspace == None


def test_build_48_well_plate():
    """
    Tests 6 plate well is built correctly

    """
    plate6 = Plate(size=48, rows=6, columns=8)
    assert plate6.size == 48
    assert plate6.rows == ["A", "B", "C", "D", "E", "F"]
    assert plate6.columns == [1, 2, 3, 4, 5, 6, 7, 8]
    assert plate6.well_volume == None
    assert plate6.deadspace == None


def test_build_96_well_plate():
    """
    Tests 6 plate well is built correctly

    """
    plate6 = Plate(size=96)
    assert plate6.size == 96
    assert plate6.rows == ["A", "B", "C", "D", "E", "F", "G", "H"]
    assert plate6.columns == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
    assert plate6.wells == [
        "A1",
        "B1",
        "C1",
        "D1",
        "E1",
        "F1",
        "G1",
        "H1",
        "A2",
        "B2",
        "C2",
        "D2",
        "E2",
        "F2",
        "G2",
        "H2",
        "A3",
        "B3",
        "C3",
        "D3",
        "E3",
        "F3",
        "G3",
        "H3",
        "A4",
        "B4",
        "C4",
        "D4",
        "E4",
        "F4",
        "G4",
        "H4",
        "A5",
        "B5",
        "C5",
        "D5",
        "E5",
        "F5",
        "G5",
        "H5",
        "A6",
        "B6",
        "C6",
        "D6",
        "E6",
        "F6",
        "G6",
        "H6",
        "A7",
        "B7",
        "C7",
        "D7",
        "E7",
        "F7",
        "G7",
        "H7",
        "A8",
        "B8",
        "C8",
        "D8",
        "E8",
        "F8",
        "G8",
        "H8",
        "A9",
        "B9",
        "C9",
        "D9",
        "E9",
        "F9",
        "G9",
        "H9",
        "A10",
        "B10",
        "C10",
        "D10",
        "E10",
        "F10",
        "G10",
        "H10",
        "A11",
        "B11",
        "C11",
        "D11",
        "E11",
        "F11",
        "G11",
        "H11",
        "A12",
        "B12",
        "C12",
        "D12",
        "E12",
        "F12",
        "G12",
        "H12",
    ]
    assert plate6.well_volume == None
    assert plate6.deadspace == None



import pytest
from platemap import Plate, add_volume, transfer


def test_add_volume():
    """
    Tests 6 plate well is built correctly

    """
    plate6 = Plate(size=6, well_volume=20000, deadspace=1000)
    add_volume(plate6, "A1", 5000, "water")
    assert plate6["A1"] == {
        "id": "A1",
        "total_volume": 5000,
        "composition": {"water": {"volume": 5000, "concentration": None}},
    }
    transfer(plate6, "A1", plate6, "B1", 2500)
    assert plate6["A1"] == {
        "id": "A1",
        "total_volume": 2500.0,
        "composition": {"water": {"volume": 2500.0, "concentration": None}},
    }
    assert plate6["B1"] == {
        "id": "B1",
        "total_volume": 2500.0,
        "composition": {"water": {"volume": 2500.0, "concentration": None}},
    }
    assert plate6.contents == {
        "A1": {
            "id": "A1",
            "total_volume": 2500.0,
            "composition": {"water": {"volume": 2500.0, "concentration": None}},
        },
        "B1": {
            "id": "B1",
            "total_volume": 2500.0,
            "composition": {"water": {"volume": 2500.0, "concentration": None}},
        },
        "A2": {"id": "A2", "total_volume": 0.0, "composition": {}},
        "B2": {"id": "B2", "total_volume": 0.0, "composition": {}},
        "A3": {"id": "A3", "total_volume": 0.0, "composition": {}},
        "B3": {"id": "B3", "total_volume": 0.0, "composition": {}},
    }


def test_transfer_source_contains_multiple_liquids():
    plate6 = Plate(size=6, well_volume=20000, deadspace=1000)
    add_volume(plate6, "A1", 6000, "water")
    add_volume(plate6, "A1", 2000, "juice")
    assert plate6.contents == {
        "A1": {
            "id": "A1",
            "total_volume": 8000.0,
            "composition": {
                "water": {"volume": 6000, "concentration": None},
                "juice": {"volume": 2000, "concentration": None},
            },
        },
        "B1": {"id": "B1", "total_volume": 0.0, "composition": {}},
        "A2": {"id": "A2", "total_volume": 0.0, "composition": {}},
        "B2": {"id": "B2", "total_volume": 0.0, "composition": {}},
        "A3": {"id": "A3", "total_volume": 0.0, "composition": {}},
        "B3": {"id": "B3", "total_volume": 0.0, "composition": {}},
    }
    transfer(plate6, "A1", plate6, "B1", 2000)
    assert plate6.contents == {
        "A1": {
            "id": "A1",
            "total_volume": 6000.0,
            "composition": {
                "water": {"volume": 4500.0, "concentration": None},
                "juice": {"volume": 1500.0, "concentration": None},
            },
        },
        "B1": {
            "id": "B1",
            "total_volume": 2000.0,
            "composition": {
                "water": {"volume": 1500.0, "concentration": None},
                "juice": {"volume": 500.0, "concentration": None},
            },
        },
        "A2": {"id": "A2", "total_volume": 0.0, "composition": {}},
        "B2": {"id": "B2", "total_volume": 0.0, "composition": {}},
        "A3": {"id": "A3", "total_volume": 0.0, "composition": {}},
        "B3": {"id": "B3", "total_volume": 0.0, "composition": {}},
    }



from ..plate import Plate
from .add_volume import add_volume
import math


def assign_source_wells(
    plate: Plate,
    input_dict: dict[str, float],
    fill_to: int = None,
    alternate_wells=False,
):
    useable_liquid_per_well = (
        fill_to - plate.deadspace if fill_to else plate.well_volume - plate.deadspace
    )
    well_index = 0
    for item in input_dict:
        for well_needed in range(
            math.ceil(
                input_dict[item]
                / (
                    fill_to - plate.deadspace
                    if fill_to
                    else plate.well_volume - plate.deadspace
                )
            )
        ):
            if not alternate_wells:
                plate.set_well_id(plate.wells[well_index], item)
                add_volume(
                    plate,
                    plate.wells[well_index],
                    fill_to if fill_to else plate.well_volume,
                    item,
                )
            else:
                plate.set_well_id(plate.wells[well_index * 2], item)
                add_volume(
                    plate,
                    plate.wells[well_index * 2],
                    fill_to if fill_to else plate.well_volume,
                    item,
                )
            well_index += 1



from ..plate import Plate


def find_well(
    target_plate: Plate,
    target_volume_id: str,
    target_volume: int,
):
    try:
        return list(
            filter(
                lambda x: target_plate[x]["composition"][target_volume_id]["volume"]
                >= (target_plate.deadspace + target_volume),
                list(
                    filter(
                        lambda x: target_volume_id in target_plate[x]["composition"],
                        target_plate.wells,
                    )
                ),
            )
        )[0]
    except:
        return 0



from ..plate import Plate


def remove_volume(
    target_plate: Plate,
    target_well: str,
    volume: int,
):
    """Remove volume from a well of a plate object

    Args:
        target_plate (Plate): target plate in which volume will be removed from,
        target_well (str): target well in which volume will be removed from,
        volume (float): volume of liquid to add to the specified well,

    Raises:
        ValueError: if volume well added passes well deadspace
    """
    if target_plate[target_well]["total_volume"] - volume >= target_plate.deadspace:
        for item in target_plate[target_well]["composition"]:
            volume_removed = (
                volume / target_plate[target_well]["total_volume"]
            ) * target_plate[target_well]["composition"][item]["volume"]
            new_volume = (
                target_plate[target_well]["composition"][item]["volume"]
                - volume_removed
            )
            target_plate.update_well_composition(target_well, new_volume, item)
        target_plate.set_well_volume(
            target_well,
            target_plate[target_well]["total_volume"] - volume,
        )
    else:
        raise ValueError(
            f"attempted to subtract volume from plate: {target_plate.id} with deadspace ({target_plate.deadspace}) "
            f"by removing {volume} nanoliters to already present {target_plate[target_well]['total_volume']} "
            f"at plate location {target_well}"
        )



import pytest
from platemap import Plate, add_volume


def test_add_volume():
    """
    Tests 6 plate well is built correctly

    """
    plate96 = Plate(size=6, well_volume=20000)
    add_volume(plate96, "A1", 5000, "water")
    assert plate96["A1"] == {
        "id": "A1",
        "total_volume": 5000,
        "composition": {"water": {"volume": 5000, "concentration": None}},
    }
    add_volume(plate96, "A1", 5000, "water")
    assert plate96["A1"] == {
        "id": "A1",
        "total_volume": 10000,
        "composition": {"water": {"volume": 10000, "concentration": None}},
    }


def test_add_multiple_volumes():
    """
    Tests 6 plate well is built correctly

    """
    plate96 = Plate(size=6, well_volume=20000)
    add_volume(plate96, "A1", 5000, "water")
    assert plate96["A1"] == {
        "id": "A1",
        "total_volume": 5000,
        "composition": {"water": {"volume": 5000, "concentration": None}},
    }
    add_volume(plate96, "A1", 2000, "juice")
    assert plate96["A1"] == {
        "id": "A1",
        "total_volume": 7000,
        "composition": {
            "water": {"volume": 5000, "concentration": None},
            "juice": {"volume": 2000, "concentration": None},
        },
    }


def test_add_too_much_volume():
    """
    Tests 6 plate well is built correctly

    """
    plate96 = Plate(size=6, well_volume=20000)
    with pytest.raises(ValueError):
        add_volume(plate96, "A1", 40000, "water")



import pytest
from platemap import Plate


def test_build_6_well_plate():
    """
    Tests 6 plate well is built correctly

    """
    plate6 = Plate(size=6)
    for index in plate6.wells:
        assert plate6[index] == {"id": index, "total_volume": 0, "composition": {}}



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="platemap",
    version="0.0.6",
    url="https://github.com/Benedict-Carling/platemap",
    author="Benedict Carling",
    author_email="benedict.carling18@imperial.ac.uk",
    license="MIT",
    description="A package for working with plates",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=("tests", "docsource")),
    include_package_data=True,
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    python_requires=">=3.7",
)



from .plate import Plate

from .PlateUtils import add_volume
from .PlateUtils import remove_volume
from .PlateUtils import transfer
from .PlateUtils import assign_source_wells
from .PlateUtils import find_well



import pytest
from platemap import Plate, assign_source_wells


def test_recurring_values_in_division():
    plate6 = Plate(size=6, well_volume=50000, deadspace=20000)
    assign_source_wells(plate6, {"water": 40000, "juice": 30000}, fill_to=40000)
    assert plate6.contents == {
        "A1": {
            "id": "water",
            "total_volume": 40000.0,
            "composition": {"water": {"volume": 40000, "concentration": None}},
        },
        "B1": {
            "id": "water",
            "total_volume": 40000.0,
            "composition": {"water": {"volume": 40000, "concentration": None}},
        },
        "A2": {
            "id": "juice",
            "total_volume": 40000.0,
            "composition": {"juice": {"volume": 40000, "concentration": None}},
        },
        "B2": {
            "id": "juice",
            "total_volume": 40000.0,
            "composition": {"juice": {"volume": 40000, "concentration": None}},
        },
        "A3": {"id": "A3", "total_volume": 0.0, "composition": {}},
        "B3": {"id": "B3", "total_volume": 0.0, "composition": {}},
    }



from ..plate import Plate
from .remove_volume import remove_volume
from .add_volume import add_volume


def transfer(
    source_plate: Plate,
    source_well: str,
    destination_plate: Plate,
    destination_well: str,
    volume: int,
):
    """Transfer volume from one well within a plate object to another well within a plate object

    Args:
        source_plate (Plate): source plate in which volume will be taken from,
        source_well (str): source well in which volume will be taken from,
        destination_plate (Plate): target plate in which volume will be added to,
        destination_well (str): target well in which volume will be added to,
        volume (float): volume of liquid to add to the specified well,

    Raises:
        ValueError: if volume in destination well added exceeds well capacity or volume in source
            well will drop below deadspace
    """
    if (
        destination_plate[destination_well]["total_volume"] + volume
        <= destination_plate.well_volume
    ):
        if source_plate[source_well]["total_volume"] - volume >= source_plate.deadspace:
            for item in source_plate[source_well]["composition"]:
                add_volume(
                    destination_plate,
                    destination_well,
                    (
                        source_plate[source_well]["composition"][item]["volume"]
                        / source_plate[source_well]["total_volume"]
                    )
                    * volume,
                    item,
                    source_plate[source_well]["composition"][item]["concentration"],
                )
            remove_volume(source_plate, source_well, volume)
        else:
            raise ValueError(
                f"attempted to subtract volume from plate: {source_plate.id} with deadspace ({source_plate.deadspace}) "
                f"by removing {volume} nanoliters to already present {source_plate[source_well]['total_volume']} "
                f"at plate location {source_well}"
            )
    else:
        raise ValueError(
            f"attempted to exceed plate: {destination_plate.id} with capacity ({destination_plate.well_volume}) "
            f"by adding {volume} nanoliters to already present {destination_plate[destination_well]['total_volume']} "
            f"at plate location {destination_well}"
        )



import pytest
from platemap import Plate, add_volume, remove_volume


def test_remove_volume():
    """
    Tests 6 plate well is built correctly

    """
    plate96 = Plate(size=6, well_volume=20000, deadspace=1000)
    add_volume(plate96, "A1", 10000, "water")
    remove_volume(plate96, "A1", 2000)
    assert plate96["A1"] == {
        "id": "A1",
        "total_volume": 8000,
        "composition": {"water": {"volume": 8000, "concentration": None}},
    }


def test_appoach_deadspace():
    """
    Tests 6 plate well is built correctly

    """
    plate96 = Plate(size=6, well_volume=20000, deadspace=1000)
    add_volume(plate96, "A1", 5000, "water")
    remove_volume(plate96, "A1", 4000)
    assert plate96["A1"] == {
        "id": "A1",
        "total_volume": 1000,
        "composition": {
            "water": {"volume": 1000, "concentration": None},
        },
    }
    with pytest.raises(ValueError):
        remove_volume(plate96, "A1", 100)


def test_division_multiple_components():
    """
    Tests 6 plate well is built correctly

    """
    plate96 = Plate(size=6, well_volume=20000, deadspace=1000)
    add_volume(plate96, "A1", 3000, "water")
    add_volume(plate96, "A1", 2000, "juice")
    remove_volume(plate96, "A1", 1000)
    assert plate96["A1"] == {
        "id": "A1",
        "total_volume": 4000,
        "composition": {
            "water": {"volume": 2400.0, "concentration": None},
            "juice": {"volume": 1600.0, "concentration": None},
        },
    }
    remove_volume(plate96, "A1", 2000)
    assert plate96["A1"] == {
        "id": "A1",
        "total_volume": 2000,
        "composition": {
            "water": {"volume": 1200.0, "concentration": None},
            "juice": {"volume": 800.0, "concentration": None},
        },
    }


def test_recurring_values_in_division():
    plate96 = Plate(size=6, well_volume=20000, deadspace=1000)
    add_volume(plate96, "A1", 2000, "water")
    add_volume(plate96, "A1", 1000, "juice")
    remove_volume(plate96, "A1", 2000)
    assert plate96["A1"] == {
        "id": "A1",
        "total_volume": 1000,
        "composition": {
            "water": {"volume": 666.6666666666667, "concentration": None},
            "juice": {"volume": 333.33333333333337, "concentration": None},
        },
    }
